/*
 * kgraft_patch_bsc1046191
 *
 * Fix for CVE-2017-7645, bsc#1046191
 *
 *  Upstream commit:
 *  e6838a29ecb484c97e4efef9429643b9851fba6e ("nfsd: check for oversized NFSv2/v3 arguments")
 *
 *  SLE12-SP2 commit:
 *  e9768725cf371b604e075deedbf11a897c403091
 *
 *  SLE12(-SP1) commit:
 *  4fb40dd8fe2396a3fe19fe00b4a5c5ea12c4b1ac
 *
 *  Copyright (c) 2017 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/sunrpc/xdr.h>
#include <linux/sunrpc/svc.h>
#include <linux/nfs.h>
#include <linux/nfs_fs.h>
#include <linux/nfsd/debug.h>

#if !IS_MODULE(CONFIG_NFSD)
#error "KGR patch supports only CONFIG_NFSD=m."
#endif

#define KGR_PATCHED_MODULE "nfsd"

/* from patched fs/nfsd/nfssvc.c, needed for dprintk() */
#define NFSDDBG_FACILITY	NFSDDBG_SVC

/*
 * From fs/nfsd/nfsd.h. Be careful to avoid a dependency to the
 * nfsd_debug exported from sunrpc.
 */
#undef ifdebug
#ifdef CONFIG_SUNRPC_DEBUG
# define ifdebug(flag)		if (*kgr_nfsd_debug & NFSDDBG_##flag)
#else
# define ifdebug(flag)		if (0)
#endif

/* from fs/nfsd/nfsd.h */
#define	kgr_nfserr_acces		cpu_to_be32(NFSERR_ACCES)
#define	kgr_nfserr_jukebox		cpu_to_be32(NFSERR_JUKEBOX)
#define	kgr_nfserr_wrongsec		cpu_to_be32(NFSERR_WRONGSEC)
#define	kgr_nfserr_dropit		cpu_to_be32(30000)

/* from fs/nfsd/cache.h */
enum {
	KGR_RC_DROPIT,
	KGR_RC_REPLY,
	KGR_RC_DOIT
};

/* from fs/nfsd/cache.h */
enum {
	KGR_RC_NOCACHE,
	KGR_RC_REPLSTAT,
	KGR_RC_REPLBUFF,
};

/* provided by fs/nfsd/nfscache.c */
static int (*kgr_nfsd_cache_lookup)(struct svc_rqst *rqstp);
static void (*kgr_nfsd_cache_update)(struct svc_rqst *rqstp,
					int cachetype, __be32 *statp);

/* provided from net/sunrpc/sysctl.c, needed by ifdebug #define above */
#ifdef CONFIG_SUNRPC_DEBUG
static unsigned int *kgr_nfsd_debug;
#endif

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "nfsd:nfsd_cache_lookup", (void *)&kgr_nfsd_cache_lookup },
	{ "nfsd:nfsd_cache_update", (void *)&kgr_nfsd_cache_update },
	{ "sunrpc:nfsd_debug", (void *)&kgr_nfsd_debug },
};


/* new, needed for the fix of nfsd_dispatch() below */
static bool kgr_nfs_request_too_big(struct svc_rqst *rqstp,
				struct svc_procedure *proc)
{
	/*
	 * The ACL code has more careful bounds-checking and is not
	 * susceptible to this problem:
	 */
	if (rqstp->rq_prog != NFS_PROGRAM)
		return false;
	/*
	 * Ditto NFSv4 (which can in theory have argument and reply both
	 * more than a page):
	 */
	if (rqstp->rq_vers >= 4)
		return false;
	/* The reply will be small, we're OK: */
	if (proc->pc_xdrressize > 0 &&
		proc->pc_xdrressize < XDR_QUADLEN(PAGE_SIZE))
		return false;

	return rqstp->rq_arg.len > PAGE_SIZE;
}

/* inlined */
static __be32 kgr_map_new_errors(u32 vers, __be32 nfserr)
{
	if (nfserr == kgr_nfserr_jukebox && vers == 2)
		return kgr_nfserr_dropit;
	if (nfserr == kgr_nfserr_wrongsec && vers < 4)
		return kgr_nfserr_acces;
	return nfserr;
}

/* patched */
int kgr_nfsd_dispatch(struct svc_rqst *rqstp, __be32 *statp)
{
	struct svc_procedure	*proc;
	kxdrproc_t		xdr;
	__be32			nfserr;
	__be32			*nfserrp;

	dprintk("nfsd_dispatch: vers %d proc %d\n",
				rqstp->rq_vers, rqstp->rq_proc);
	proc = rqstp->rq_procinfo;

	/*
	 * Fix CVE-2017-7645
	 *  +5 lines
	 */
	if (kgr_nfs_request_too_big(rqstp, proc)) {
		dprintk("nfsd: NFSv%d argument too large\n", rqstp->rq_vers);
		*statp = rpc_garbage_args;
		return 1;
	}
	/*
	 * Give the xdr decoder a chance to change this if it wants
	 * (necessary in the NFSv4.0 compound case)
	 */
	rqstp->rq_cachetype = proc->pc_cachetype;
	/* Decode arguments */
	xdr = proc->pc_decode;
	if (xdr && !xdr(rqstp, (__be32*)rqstp->rq_arg.head[0].iov_base,
			rqstp->rq_argp)) {
		dprintk("nfsd: failed to decode arguments!\n");
		*statp = rpc_garbage_args;
		return 1;
	}

	/* Check whether we have this call in the cache. */
	switch (kgr_nfsd_cache_lookup(rqstp)) {
	case KGR_RC_DROPIT:
		return 0;
	case KGR_RC_REPLY:
		return 1;
	case KGR_RC_DOIT:;
		/* do it */
	}

	/* need to grab the location to store the status, as
	 * nfsv4 does some encoding while processing
	 */
	nfserrp = rqstp->rq_res.head[0].iov_base
		+ rqstp->rq_res.head[0].iov_len;
	rqstp->rq_res.head[0].iov_len += sizeof(__be32);

	/* Now call the procedure handler, and encode NFS status. */
	nfserr = proc->pc_func(rqstp, rqstp->rq_argp, rqstp->rq_resp);
	nfserr = kgr_map_new_errors(rqstp->rq_vers, nfserr);
	if (nfserr == kgr_nfserr_dropit || test_bit(RQ_DROPME, &rqstp->rq_flags)) {
		dprintk("nfsd: Dropping request; may be revisited later\n");
		kgr_nfsd_cache_update(rqstp, KGR_RC_NOCACHE, NULL);
		return 0;
	}

	if (rqstp->rq_proc != 0)
		*nfserrp++ = nfserr;

	/* Encode result.
	 * For NFSv2, additional info is never returned in case of an error.
	 */
	if (!(nfserr && rqstp->rq_vers == 2)) {
		xdr = proc->pc_encode;
		if (xdr && !xdr(rqstp, nfserrp,
				rqstp->rq_resp)) {
			/* Failed to encode result. Release cache entry */
			dprintk("nfsd: failed to encode result!\n");
			kgr_nfsd_cache_update(rqstp, KGR_RC_NOCACHE, NULL);
			*statp = rpc_system_err;
			return 1;
		}
	}

	/* Store reply in cache. */
	kgr_nfsd_cache_update(rqstp, rqstp->rq_cachetype, statp + 1);
	return 1;
}

static int kgr_patch_bsc1046191_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1046191_module_notify(struct notifier_block *nb,
	unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1046191_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1046191_module_nb = {
	.notifier_call = kgr_patch_bsc1046191_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1046191_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1046191_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1046191_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1046191_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1046191_module_nb);
}
